from __future__ import annotations

import datetime
import uuid
from typing import Any

from pydantic import BaseModel, Field

import ragna.core


class Components(BaseModel):
    documents: list[str]
    source_storages: list[dict[str, Any]]
    assistants: list[dict[str, Any]]


class Document(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    name: str

    @classmethod
    def from_core(cls, document: ragna.core.Document) -> Document:
        return cls(
            id=document.id,
            name=document.name,
        )


class DocumentUploadInfo(BaseModel):
    url: str
    data: dict
    document: Document


class Source(BaseModel):
    # See orm.Source on why this is not a UUID
    id: str
    document: Document
    location: str

    @classmethod
    def from_core(cls, source: ragna.core.Source) -> Source:
        return cls(
            id=source.id,
            document=Document.from_core(source.document),
            location=source.location,
        )


class Message(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    content: str
    role: ragna.core.MessageRole
    sources: list[Source] = Field(default_factory=list)
    timestamp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.utcnow()
    )

    @classmethod
    def from_core(cls, message: ragna.core.Message) -> Message:
        return cls(
            content=message.content,
            role=message.role,
            sources=[Source.from_core(source) for source in message.sources],
        )


class ChatMetadata(BaseModel):
    name: str
    source_storage: str
    assistant: str
    params: dict
    documents: list[Document]


class Chat(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    metadata: ChatMetadata
    messages: list[Message] = Field(default_factory=list)
    prepared: bool = False


class MessageOutput(BaseModel):
    message: Message
    chat: Chat



import panel as pn
import param

from ragna import __version__ as ragna_version


class LeftSidebar(pn.viewable.Viewer):
    chats = param.List(default=[])
    current_chat_id = param.String(default=None)
    refresh_counter = param.Integer(default=0)

    def __init__(self, api_wrapper, **params):
        super().__init__(**params)

        self.on_click_chat = None
        self.on_click_new_chat = None

        self.chat_buttons = []

        pn.state.location.sync(
            self,
            {"current_chat_id": "current_chat_id"},
            on_error=lambda x: print(f"error sync on {x}"),
        )

    def trigger_on_click_new_chat(self, event):
        if event.old > event.new:
            return

        if self.on_click_new_chat is not None:
            self.on_click_new_chat(event)

    def on_click_chat_wrapper(self, event, chat):
        # This is a hack to avoid the event being triggered twice in a row
        if event.old > event.new:
            return

        # update the UI, unselect all buttons ...
        for button in self.chat_buttons:
            if "selected" in button.css_classes:
                button.css_classes = []

        # ... and select the one that was clicked
        event.obj.css_classes = ["selected"]

        # call the actual callback
        if self.on_click_chat is not None:
            self.on_click_chat(chat)

    def footer(self):
        return pn.pane.HTML(
            """<div class='user_block'> Username </div>
                            """,
            styles={},
            stylesheets=[
                """ 
                                            
                                :host {
                                        height: 64px;
                                        width: 100%;
                                        margin: 0;
                                    
                                }

                                :host .user_block {
                                            height:64px;
                                            background-color: rgb(248, 248, 248);
                                            border-top: solid 1px rgb(248, 248, 248);
                                }
                            """
            ],
        )

    def refresh(self):
        self.refresh_counter += 1

    @pn.depends("refresh_counter", "chats", "current_chat_id", on_init=True)
    def __panel__(self):
        self.chat_buttons = []
        for chat in self.chats:
            button = pn.widgets.Button(
                name=chat["metadata"]["name"], button_style="outline"
            )
            button.on_click(lambda event, c=chat: self.on_click_chat_wrapper(event, c))

            button.stylesheets.append(
                """
                                      :host {  
                                                width:90%;
                                                min-width: 200px;
                                        }
                                      
                                      :host div button {
                                        overflow: hidden;
                                        text-overflow: ellipsis;
                                        text-align:left;
                                        border: 0px !important;
                                      
                                      }

                                      :host div button:before {
                                        content: url("/imgs/chat_bubble.svg");
                                        margin-right: 10px;
                                        display: inline-block;
                                      }

                                      :host(.selected) div button, :host div button:hover {
                                        background-color: #F3F3F3 !important;
                                        border-radius: 0px 5px 5px 0px !important;
                                        border-left: solid 4px var(--accent-color) !important;
                                      }

                                      
                                      """
            )
            self.chat_buttons.append(button)

            try:
                if chat["id"] == self.current_chat_id:
                    button.css_classes = ["selected"]
            except Exception:
                pass

        header = pn.pane.HTML(
            """<img src="imgs/ragna_logo.svg" height="32px" /><span>Ragna</span>""",
            stylesheets=[
                """ 
                                               :host { 
                                                    background-color: #F9F9F9;
                                                    border-bottom: 1px solid #EEEEEE;

                                                    width: 100%;
                                                    height: 54px;
                                                    margin: 0;
                                               }

                                                :host div {
                                                    display: flex;
                                                    align-items: center;
                                                    height: 100%;
                                                }

                                                :host img {
                                                    margin: 5px;
                                                    margin-left: 12px;
                                                }

                                               :host span { 
                                                    margin-left: 20px;
                                                    font-size: 24px;
                                                    font-weight: 600;
                                               }
                                    """
            ],
        )

        new_chat_button = pn.widgets.Button(
            name="New Chat",
            button_type="primary",
            icon="plus",
            stylesheets=[
                """ 
                        :host { 
                            width: 90%;
                            margin-left: 10px;
                            margin-top: 10px;
                        }
                        :host div button { 
                            background-color: var(--accent-color) !important;
                            text-align: left;
                        }
                """
            ],
        )

        new_chat_button.on_click(self.trigger_on_click_new_chat)

        objects = (
            [header, new_chat_button]
            + self.chat_buttons
            + [
                pn.layout.VSpacer(),
                pn.pane.HTML(f"version: {ragna_version}"),
                # self.footer()
            ]
        )

        result = pn.Column(
            *objects,
            stylesheets=[
                """   
                        :host { 
                            overflow-x: hidden;
                            height: 100%;
                            width:100%;
                            border-right: 1px solid #EEEEEE;
                        }
                """
            ],
        )

        return result



import json
import uuid

import param
from panel.reactive import ReactiveHTML
from panel.widgets import Widget


class FileUploader(ReactiveHTML, Widget):  # type: ignore[misc]
    allowed_documents = param.List(default=[])
    allowed_documents_str = param.String(default="")

    file_list = param.List(default=[])

    custom_js = param.String(default="")
    uploaded_documents_json = param.String(default="")

    title = param.String(default="")

    def __init__(self, allowed_documents, token, informations_endpoint, **params):
        super().__init__(**params)

        self.token = token
        self.informations_endpoint = informations_endpoint

        self.after_upload_callback = None

    def can_proceed_to_upload(self):
        return len(self.file_list) > 0

    @param.depends("allowed_documents", watch=True)
    def update_allowed_documents_str(self):
        if len(self.allowed_documents) == 1:
            self.allowed_documents_str = (
                "Only " + self.allowed_documents[0] + " files are allowed."
            )
        else:
            self.allowed_documents_str = "Allowed files : " + ", ".join(
                self.allowed_documents
            )

    @param.depends("uploaded_documents_json", watch=True)
    async def did_finish_upload(self):
        if self.after_upload_callback is not None:
            await self.after_upload_callback(json.loads(self.uploaded_documents_json))

    def perform_upload(self, event=None, after_upload_callback=None):
        self.after_upload_callback = after_upload_callback

        self.loading = True

        final_callback_js = """
            var final_callback = function(uploaded_documents) {
                self.get_uploaded_documents_json().innerText = JSON.stringify(uploaded_documents);
            };
        """

        # This is a hack to force the re-execution of the javascript.
        # If the whole javascript is the same, and doesn't change,
        # the panel widget is not re-renderer, and the upload function is not called.
        random_id = f" var random_id = '{str(uuid.uuid4())}';"

        self.custom_js = (
            final_callback_js
            + random_id
            + f"""upload( self.get_upload_files(),  '{self.token}', '{self.informations_endpoint}', final_callback) """
        )

    _child_config = {
        "custom_js": "template",
        "uploaded_documents_json": "template",
        "allowed_documents_str": "literal",
    }

    _template = """
            <style>
                :host {
                    width: 100%;
                    margin: 0px;
                    padding: 0px;
                }

                .fileUploadContainer {
                    height: 130px;
                }

                .fileUploadDropArea {
                    margin-left: 10px;
                    margin-right: 10px;
                    border: 1px dashed var(--accent-color);
                    border-radius: 5px;
                    height:100%;
                    display:flex;
                    text-align: center;
                    justify-content: center;
                    align-items: center;
                    flex-direction: column;
                }

                .fileUploadDropArea span.bold {
                    font-weight: bold;
                }

                .fileUploadDropArea img {
                    padding-bottom:5px;
                }

                .fileUploadDropArea.draggedOver {
                    border-width: 3px;
                }

                .fileUploadDropArea.uploaded {
                    height: calc(100% - 40px);
                }
                
                .fileUpload {
                    height: 100% !important;
                    position: absolute;
                    opacity: 0;
                }

                .fileListContainer {
                    display:flex;
                    flex-direction: row;
                    height: 44px;
                    overflow:scroll;
                    padding-top:14px;
                    padding-left:6px;
                }
                
                .chat_document_pill {
                    background-color: rgb(241,241,241);
                    margin-top:0px; 
                    margin-left: 5px;   
                    margin-right: 5px;
                    padding: 5px 15px;
                    border-radius: 10px;
                    color: var(--accent-color);
                    display: inline-table;
                    }
                
            </style>
            <script>
                                                                    
                var scr = document.createElement("script");
                scr.src = "/resources/upload.js" + "?ts=" + new Date().getTime();
                document.getElementsByTagName("head")[0].appendChild(scr);

            </script>
            <div id="fileUploadContainer" class="fileUploadContainer">
                <div id="fileUploadDropArea" class="fileUploadDropArea">
                    <img src="/imgs/cloud-upload.svg" width="24px" height="24px" />
                    <span><b>Click to upload</b> or drag and drop.<br /></span>
                    <div id='allowedDocuments'>
                        ${allowed_documents_str}
                    </div>
                    <input  type="file" 
                            name="fileUpload"
                            class="fileUpload" 
                            id="fileUpload" 
                            multiple="multiple" 
                            onchange="${script('file_input_on_change')}" /> 
                </div>
                <div id="fileListContainer" class="fileListContainer">
                    
                </div>
            </div>
            <div style="display:none;">
                <div id="custom_js_watcher">
                ${custom_js}
                </div>
                <div id="uploaded_documents_json_watcher">
                    
                </div>
            </div>
    """

    _scripts = {
        "after_layout": """ 
            self.update_layout();
        """,
        "update_layout": """ 

            
            if (data.file_list.length > 0) {
                fileUploadDropArea.classList.add("uploaded");
            } else {
                fileUploadDropArea.classList.remove("uploaded");
            }

            fileListContainer.innerHTML = "";

            data.file_list.forEach(function(f) {
                var pill = document.createElement("div");
                pill.classList.add("chat_document_pill");
                var fname = document.createTextNode(f.name);

                pill.appendChild(fname);
                fileListContainer.appendChild(pill);

            });

        """,
        "file_input_on_change": """
            
            
            var new_file_list = Array.from(fileUpload.files).map(function(f) {
                new_f = {
                    "lastModified":f.lastModified ,
                    "name":f.name ,
                    "size":f.size ,
                    "type":f.type ,
                };

                return new_f;
            });
            
            data.file_list = new_file_list;
            self.update_layout();
           
            
        """,
        "get_upload_files": """
            
            return fileUpload.files;
        """,
        "get_uploaded_documents_json": """
            return uploaded_documents_json_watcher;
        """,
        "render": """
            
            var MutationObserver = window.MutationObserver || window.WebKitMutationObserver || window.MozMutationObserver;
            var observer = new MutationObserver(function(mutationsList, observer) {
                    mutationsList.forEach(function(mutation){
                        if (mutation.type == 'characterData') {
                            eval(custom_js_watcher.innerText);
                        }
                        
                    });    

            });
            observer.observe(custom_js_watcher, {characterData: true, childList: true, attributes: true, subtree: true});


            var MutationObserver = window.MutationObserver || window.WebKitMutationObserver || window.MozMutationObserver;
            var observer = new MutationObserver(function(mutationsList, observer) {
                    mutationsList.forEach(function(mutation){
                        data.uploaded_documents_json = uploaded_documents_json_watcher.innerText
                    });    

            });
            observer.observe(uploaded_documents_json_watcher, {characterData: true, childList: true, attributes: true, subtree: true});


            fileUpload.addEventListener("dragenter", function(event){
                    fileUploadDropArea.classList.add("draggedOver");
            });

            fileUpload.addEventListener("dragleave", function(event){
                    fileUploadDropArea.classList.remove("draggedOver")
            });

            fileUpload.addEventListener("drop", function(event){
                    fileUploadDropArea.classList.remove("draggedOver")
                    event.preventDefault();
                    fileUpload.files = event.dataTransfer.files;
                    self.file_input_on_change();
            });

        """,
        "remove": """  
                  
         """,
    }



"""
The tests in this module are only here as a reminder to clean up our code if an issue is
fixed upstream. If you see a test failing here, i.e. an unexpected success, feel free to
remove the offending test after you have cleaned up our code.
"""

import subprocess
import sys

import pytest

from ragna._compat import importlib_metadata_package_distributions


@pytest.mark.xfail
def test_pyarrow_dummy_module():
    module_names = {
        module_name
        for module_name, distribution_names in importlib_metadata_package_distributions().items()
        if "pyarrow" in distribution_names
    }
    assert "__dummy__" not in module_names


@pytest.mark.xfail
def test_pymupdf_import_warning():
    # See https://github.com/Quansight/ragna/issues/75
    result = subprocess.run(
        [sys.executable, "-c", "import fitz_new"], capture_output=True
    )
    assert not result.stdout
    assert not result.stderr



import panel as pn
import param

from . import js
from . import styles as ui


class ModalWelcome(pn.viewable.Viewer):
    close_button_callback = param.Callable()

    def __init__(self, **params):
        super().__init__(**params)

    def did_click_on_close_button(self, event):
        if self.close_button_callback is not None:
            self.close_button_callback()

    def __panel__(self):
        close_button = pn.widgets.Button(
            name="Okay, let's go",
            button_type="primary",
            stylesheets=[""" :host { width:35%; margin-left:60%; }"""],
        )
        close_button.on_click(self.did_click_on_close_button)

        return pn.Column(
            pn.pane.HTML(
                f"""<script>{js.reset_modal_size(ui.WELCOME_MODAL_WIDTH, ui.WELCOME_MODAL_HEIGHT)}</script>"""
                + """<h2>Welcome !</h2><br />
                        Ragna is a RAG Orchestration Framework.<br />
                        With its UI, select and configure LLMs, upload documents, and chat with the LLM.<br />
                        <br />
                        Use Ragna UI out-of-the-box, as a daily-life interface with your favorite AI, <br />
                        or as a reference to build custom web applications.
                        <br /><br /><br />
                """
            ),
            close_button,
            width=ui.WELCOME_MODAL_WIDTH,
            height=ui.WELCOME_MODAL_HEIGHT,
            sizing_mode="fixed",
        )



import contextlib
import os

import httpx
import pytest

from ragna import Config
from ragna.core._utils import default_user
from tests.utils import (
    ragna_api,
    ragna_worker,
    redis_server,
    skip_redis_on_windows,
    timeout_after,
)


@pytest.mark.parametrize(
    "queue",
    ["memory", "file_system", pytest.param("redis", marks=skip_redis_on_windows)],
)
@pytest.mark.parametrize("database", ["memory", "sqlite"])
def test_e2e(tmp_path, queue, database):
    if queue == "memory":
        queue_cm = contextlib.nullcontext("memory")
        worker_cm_fn = contextlib.nullcontext
    elif queue == "file_system":
        queue_cm = contextlib.nullcontext(str(tmp_path / "queue"))
        worker_cm_fn = ragna_worker
    elif queue == "redis":
        queue_cm = redis_server()
        worker_cm_fn = ragna_worker

    if database == "memory":
        database_url = "memory"
    elif database == "sqlite":
        database_url = f"sqlite:///{tmp_path / 'ragna.db'}"

    with queue_cm as queue_url:
        config = Config(
            local_cache_root=tmp_path,
            core=dict(queue_url=queue_url),
            api=dict(database_url=database_url),
        )
        with worker_cm_fn(config):
            check_api(config)


@timeout_after()
def check_api(config):
    document_root = config.local_cache_root / "documents"
    document_root.mkdir()
    document_path = document_root / "test.txt"
    with open(document_path, "w") as file:
        file.write("!\n")

    with ragna_api(config, start_worker=False), httpx.Client(
        base_url=config.api.url
    ) as client:
        username = default_user()
        token = (
            client.post(
                "/token",
                data={
                    "username": username,
                    "password": os.environ.get(
                        "AI_PROXY_DEMO_AUTHENTICATION_PASSWORD", username
                    ),
                },
            )
            .raise_for_status()
            .json()
        )
        client.headers["Authorization"] = f"Bearer {token}"

        assert client.get("/chats").raise_for_status().json() == []

        document_info = (
            client.get("/document", params={"name": document_path.name})
            .raise_for_status()
            .json()
        )
        document = document_info["document"]
        assert document["name"] == document_path.name

        with open(document_path, "rb") as file:
            client.post(
                document_info["url"],
                data=document_info["data"],
                files={"file": file},
            ).raise_for_status()

        components = client.get("/components").raise_for_status().json()
        documents = components["documents"]
        assert set(documents) == config.core.document.supported_suffixes()
        source_storages = [
            json_schema["title"] for json_schema in components["source_storages"]
        ]
        assert source_storages == [
            source_storage.display_name()
            for source_storage in config.core.source_storages
        ]
        assistants = [json_schema["title"] for json_schema in components["assistants"]]
        assert assistants == [
            assistant.display_name() for assistant in config.core.assistants
        ]

        source_storage = source_storages[0]
        assistant = assistants[0]

        chat_metadata = {
            "name": "test-chat",
            "source_storage": source_storage,
            "assistant": assistant,
            "params": {},
            "documents": [document],
        }
        chat = client.post("/chats", json=chat_metadata).raise_for_status().json()
        assert chat["metadata"] == chat_metadata
        assert not chat["prepared"]
        assert chat["messages"] == []

        assert client.get("/chats").raise_for_status().json() == [chat]
        assert client.get(f"/chats/{chat['id']}").raise_for_status().json() == chat

        json = client.post(f"/chats/{chat['id']}/prepare").raise_for_status().json()
        message, chat = json["message"], json["chat"]
        assert message["role"] == "system"
        assert message["sources"] == []
        assert chat["prepared"]
        assert len(chat["messages"]) == 1
        assert chat["messages"][-1] == message

        prompt = "?"
        json = (
            client.post(f"/chats/{chat['id']}/answer", params={"prompt": prompt})
            .raise_for_status()
            .json()
        )
        message, chat = json["message"], json["chat"]
        assert message["role"] == "assistant"
        assert {source["document"]["name"] for source in message["sources"]} == {
            document_path.name
        }
        assert len(chat["messages"]) == 3
        assert (
            chat["messages"][-2]["role"] == "user"
            and chat["messages"][-2]["sources"] == []
            and chat["messages"][-2]["content"] == prompt
        )
        assert chat["messages"][-1] == message

        client.delete(f"/chats/{chat['id']}").raise_for_status()
        assert client.get("/chats").raise_for_status().json() == []



import asyncio

import pydantic
import pytest

from ragna import Rag, assistants, source_storages


@pytest.fixture
def demo_document(tmp_path, request):
    path = tmp_path / "demo_document.txt"
    with open(path, "w") as file:
        file.write(f"{request.node.name}\n")
    return path


def test_chat_params_extra(demo_document):
    async def main():
        async with Rag().chat(
            documents=[demo_document],
            source_storage=source_storages.RagnaDemoSourceStorage,
            assistant=assistants.RagnaDemoAssistant,
            not_supported_parameter="arbitrary_value",
        ):
            pass

    with pytest.raises(pydantic.ValidationError, match="not_supported_parameter"):
        asyncio.run(main())



import datetime
import json
import os

import httpx

from ragna.core._utils import default_user


def main():
    client = httpx.Client(base_url="http://127.0.0.1:31476")
    client.get("/").raise_for_status()

    ## authentication

    username = default_user()
    token = (
        client.post(
            "/token",
            data={
                "username": username,
                "password": os.environ.get(
                    "AI_PROXY_DEMO_AUTHENTICATION_PASSWORD", username
                ),
            },
        )
        .raise_for_status()
        .json()
    )
    client.headers["Authorization"] = f"Bearer {token}"

    ## documents

    documents = []
    for i in range(5):
        name = f"document{i}.txt"
        document_info = (
            client.get("/document", params={"name": name}).raise_for_status().json()
        )
        client.post(
            document_info["url"],
            data=document_info["data"],
            files={"file": f"Content of {name}".encode()},
        ).raise_for_status()
        documents.append(document_info["document"])

    ## chat 1

    chat = (
        client.post(
            "/chats",
            json={
                "name": "Test chat",
                "documents": documents[:2],
                "source_storage": "Ragna/DemoSourceStorage",
                "assistant": "Ragna/DemoAssistant",
                "params": {},
            },
        )
        .raise_for_status()
        .json()
    )

    client.post(f"/chats/{chat['id']}/prepare").raise_for_status()
    client.post(
        f"/chats/{chat['id']}/answer",
        params={"prompt": "Hello!"},
    ).raise_for_status()

    ## chat 2

    chat = (
        client.post(
            "/chats",
            json={
                "name": f"Chat {datetime.datetime.now():%x %X}",
                "documents": documents[2:4],
                "source_storage": "Ragna/DemoSourceStorage",
                "assistant": "Ragna/DemoAssistant",
                "params": {},
            },
        )
        .raise_for_status()
        .json()
    )
    client.post(f"/chats/{chat['id']}/prepare").raise_for_status()
    for _ in range(3):
        client.post(
            f"/chats/{chat['id']}/answer",
            params={"prompt": "What is Ragna? Please, I need to know!"},
        ).raise_for_status()

    ## chat 3

    chat = (
        client.post(
            "/chats",
            json={
                "name": (
                    "Really long chat name that likely needs to be truncated somehow. "
                    "If you can read this, truncating failed :boom:"
                ),
                "documents": [documents[i] for i in [0, 2, 4]],
                "source_storage": "Ragna/DemoSourceStorage",
                "assistant": "Ragna/DemoAssistant",
                "params": {},
            },
        )
        .raise_for_status()
        .json()
    )
    client.post(f"/chats/{chat['id']}/prepare").raise_for_status()
    client.post(
        f"/chats/{chat['id']}/answer",
        params={"prompt": "Hello!"},
    ).raise_for_status()
    client.post(
        f"/chats/{chat['id']}/answer",
        params={"prompt": "Ok, in that case show me some pretty markdown!"},
    ).raise_for_status()

    chats = client.get("/chats").raise_for_status().json()
    print(json.dumps(chats))


if __name__ == "__main__":
    main()



import contextlib
import uuid
from typing import Annotated, Any, Iterator, Type, cast

import aiofiles
from fastapi import Depends, FastAPI, Form, HTTPException, Request, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

import ragna
import ragna.core
from ragna._utils import handle_localhost_origins
from ragna.core import Config, Rag, RagnaException
from ragna.core._components import Component
from ragna.core._rag import SpecialChatParams

from . import database, schemas


def app(config: Config) -> FastAPI:
    rag = Rag(config)

    app = FastAPI(title="ragna", version=ragna.__version__)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=handle_localhost_origins(config.api.origins),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.exception_handler(RagnaException)
    async def ragna_exception_handler(
        request: Request, exc: RagnaException
    ) -> JSONResponse:
        if exc.http_detail is RagnaException.EVENT:
            detail = exc.event
        elif exc.http_detail is RagnaException.MESSAGE:
            detail = str(exc)
        else:
            detail = cast(str, exc.http_detail)
        return JSONResponse(
            status_code=exc.http_status_code,
            content={"error": {"message": detail}},
        )

    @app.get("/")
    async def version() -> str:
        return ragna.__version__

    authentication = config.api.authentication()

    @app.post("/token")
    async def create_token(request: Request) -> str:
        return await authentication.create_token(request)

    UserDependency = Annotated[str, Depends(authentication.get_user)]

    def _get_component_json_schema(
        component: Type[Component],
    ) -> dict[str, dict[str, Any]]:
        json_schema = component._protocol_model().model_json_schema()
        # FIXME: there is likely a better way to exclude certain fields builtin in
        #  pydantic
        for special_param in SpecialChatParams.model_fields:
            if (
                "properties" in json_schema
                and special_param in json_schema["properties"]
            ):
                del json_schema["properties"][special_param]
            if "required" in json_schema and special_param in json_schema["required"]:
                json_schema["required"].remove(special_param)
        return json_schema

    @app.get("/components")
    async def get_components(_: UserDependency) -> schemas.Components:
        return schemas.Components(
            documents=sorted(config.core.document.supported_suffixes()),
            source_storages=[
                _get_component_json_schema(source_storage)
                for source_storage in config.core.source_storages
            ],
            assistants=[
                _get_component_json_schema(assistant)
                for assistant in config.core.assistants
            ],
        )

    database_url = config.api.database_url
    if database_url == "memory":
        database_url = "sqlite://"
    make_session = database.get_sessionmaker(database_url)

    @contextlib.contextmanager
    def get_session() -> Iterator[database.Session]:
        with make_session() as session:  # type: ignore[attr-defined]
            yield session

    @app.get("/document")
    async def get_document_upload_info(
        user: UserDependency,
        name: str,
    ) -> schemas.DocumentUploadInfo:
        with get_session() as session:
            document = schemas.Document(name=name)
            url, data, metadata = await config.core.document.get_upload_info(
                config=config, user=user, id=document.id, name=document.name
            )
            database.add_document(
                session, user=user, document=document, metadata=metadata
            )
            return schemas.DocumentUploadInfo(url=url, data=data, document=document)

    @app.post("/document")
    async def upload_document(
        token: Annotated[str, Form()], file: UploadFile
    ) -> schemas.Document:
        if not issubclass(rag.config.core.document, ragna.core.LocalDocument):
            raise HTTPException(
                status_code=400,
                detail="Ragna configuration does not support local upload",
            )
        with get_session() as session:
            user, id = ragna.core.LocalDocument.decode_upload_token(token)
            document, metadata = database.get_document(session, user=user, id=id)

            core_document = ragna.core.LocalDocument(
                id=document.id, name=document.name, metadata=metadata
            )
            core_document.path.parent.mkdir(parents=True, exist_ok=True)
            async with aiofiles.open(core_document.path, "wb") as document_file:
                while content := await file.read(1024):
                    await document_file.write(content)

            return document

    def schema_to_core_chat(
        session: database.Session, *, user: str, chat: schemas.Chat
    ) -> ragna.core.Chat:
        core_chat = rag.chat(
            documents=[
                rag.config.core.document(
                    id=document.id,
                    name=document.name,
                    metadata=database.get_document(
                        session,
                        user=user,
                        id=document.id,
                    )[1],
                )
                for document in chat.metadata.documents
            ],
            source_storage=chat.metadata.source_storage,
            assistant=chat.metadata.assistant,
            user=user,
            chat_id=chat.id,
            chat_name=chat.metadata.name,
            **chat.metadata.params,
        )
        # FIXME: We need to reconstruct the previous messages here. Right now this is
        #  not needed, because the chat itself never accesses past messages. However,
        #  if we implement a chat history feature, i.e. passing past messages to
        #  the assistant, this becomes crucial.
        core_chat._messages = []
        core_chat._prepared = chat.prepared

        return core_chat

    @app.post("/chats")
    async def create_chat(
        user: UserDependency,
        chat_metadata: schemas.ChatMetadata,
    ) -> schemas.Chat:
        with get_session() as session:
            chat = schemas.Chat(metadata=chat_metadata)

            # Although we don't need the actual ragna.core.Chat object here,
            # we use it to validate the documents and metadata.
            schema_to_core_chat(session, user=user, chat=chat)

            database.add_chat(session, user=user, chat=chat)
            return chat

    @app.get("/chats")
    async def get_chats(user: UserDependency) -> list[schemas.Chat]:
        with get_session() as session:
            return database.get_chats(session, user=user)

    @app.get("/chats/{id}")
    async def get_chat(user: UserDependency, id: uuid.UUID) -> schemas.Chat:
        with get_session() as session:
            return database.get_chat(session, user=user, id=id)

    @app.post("/chats/{id}/prepare")
    async def prepare_chat(
        user: UserDependency, id: uuid.UUID
    ) -> schemas.MessageOutput:
        with get_session() as session:
            chat = database.get_chat(session, user=user, id=id)

            core_chat = schema_to_core_chat(session, user=user, chat=chat)
            welcome = schemas.Message.from_core(await core_chat.prepare())
            chat.prepared = True
            chat.messages.append(welcome)

            database.update_chat(session, user=user, chat=chat)

            return schemas.MessageOutput(message=welcome, chat=chat)

    @app.post("/chats/{id}/answer")
    async def answer(
        user: UserDependency, id: uuid.UUID, prompt: str
    ) -> schemas.MessageOutput:
        with get_session() as session:
            chat = database.get_chat(session, user=user, id=id)
            chat.messages.append(
                schemas.Message(content=prompt, role=ragna.core.MessageRole.USER)
            )

            core_chat = schema_to_core_chat(session, user=user, chat=chat)

            answer = schemas.Message.from_core(await core_chat.answer(prompt))
            chat.messages.append(answer)

            database.update_chat(session, user=user, chat=chat)

            return schemas.MessageOutput(message=answer, chat=chat)

    @app.delete("/chats/{id}")
    async def delete_chat(user: UserDependency, id: uuid.UUID) -> None:
        with get_session() as session:
            database.delete_chat(session, user=user, id=id)

    return app



from __future__ import annotations

import abc
import contextlib
import enum
import functools
import getpass
import importlib
import importlib.metadata
import os
from collections import defaultdict
from typing import Any, Collection, Optional, Type, Union, cast

import packaging.requirements
import pydantic
import pydantic_core

from ragna._compat import importlib_metadata_package_distributions


class RagnaExceptionHttpDetail(enum.Enum):
    EVENT = enum.auto()
    MESSAGE = enum.auto()


class RagnaException(Exception):
    """Ragna exception."""

    # The values below are sentinels to be used with the http_detail field.
    # They tells the API to use the event as detail in the returned error message
    EVENT = RagnaExceptionHttpDetail.EVENT
    MESSAGE = RagnaExceptionHttpDetail.MESSAGE

    def __init__(
        self,
        # FIXME: remove default value for event
        event: str = "",
        http_status_code: int = 500,
        http_detail: Union[str, RagnaExceptionHttpDetail] = "",
        **extra: Any,
    ) -> None:
        self.event = event
        self.http_status_code = http_status_code
        self.http_detail = http_detail
        self.extra = extra

    def __str__(self) -> str:
        return ", ".join([self.event, *[f"{k}={v}" for k, v in self.extra.items()]])


class Requirement(abc.ABC):
    @abc.abstractmethod
    def is_available(self) -> bool:
        ...

    @abc.abstractmethod
    def __repr__(self) -> str:
        ...

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Requirement):
            return NotImplemented

        return repr(self) == repr(other)


class RequirementsMixin:
    @classmethod
    def requirements(cls) -> list[Requirement]:
        return []

    @classmethod
    def is_available(cls) -> bool:
        return all(requirement.is_available() for requirement in cls.requirements())


class PackageRequirement(Requirement):
    def __init__(
        self, requirement_string: str, *, exclude_modules: Collection[str] = ()
    ) -> None:
        self._requirement = packaging.requirements.Requirement(requirement_string)
        self._exclude_modules = set(exclude_modules)

    @functools.cache
    def is_available(self) -> bool:
        try:
            distribution = importlib.metadata.distribution(self._requirement.name)
        except importlib.metadata.PackageNotFoundError:
            return False

        if distribution.version not in self._requirement.specifier:
            return False

        for module_name in {
            module_name
            for module_name, distribution_names in importlib_metadata_package_distributions().items()
            if distribution.name in distribution_names  # type: ignore[attr-defined]
            and module_name not in self._exclude_modules
        }:
            try:
                importlib.import_module(module_name)
            except Exception:
                return False

        return True

    def __repr__(self) -> str:
        return str(self._requirement)


class EnvVarRequirement(Requirement):
    def __init__(self, name: str) -> None:
        self._name = name

    @functools.cache
    def is_available(self) -> bool:
        return self._name in os.environ

    def __repr__(self) -> str:
        return self._name


def default_user() -> str:
    with contextlib.suppress(Exception):
        return getpass.getuser()
    with contextlib.suppress(Exception):
        return os.getlogin()
    return "Ragna"


def merge_models(
    model_name: str,
    *models: Type[pydantic.BaseModel],
    config: Optional[pydantic.ConfigDict] = None,
) -> Type[pydantic.BaseModel]:
    raw_field_definitions = defaultdict(list)
    for model_cls in models:
        for name, field in model_cls.model_fields.items():
            type_ = field.annotation

            default: Any
            if field.is_required():
                default = ...
            elif field.default is pydantic_core.PydanticUndefined:
                default = field.default_factory()  # type: ignore[misc]
            else:
                default = field.default

            raw_field_definitions[name].append((type_, default))

    field_definitions = {}
    for name, definitions in raw_field_definitions.items():
        types, defaults = zip(*definitions)

        types = set(types)
        if len(types) > 1:
            raise RagnaException(f"Mismatching types for field {name}: {types}")
        type_ = types.pop()

        defaults = set(defaults)
        if len(defaults) == 1:
            default = defaults.pop()
        elif ... in defaults:
            default = ...
        else:
            default = None

        field_definitions[name] = (type_, default)

    return cast(
        Type[pydantic.BaseModel],
        pydantic.create_model(  # type: ignore[call-overload]
            model_name, **field_definitions, __config__=config
        ),
    )



from typing import cast

from ragna.core import RagnaException, Source

from ._api import ApiAssistant


class MosaicmlApiAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "MOSAICML_API_KEY"
    _MODEL: str
    _CONTEXT_SIZE: int

    @classmethod
    def display_name(cls) -> str:
        return f"MosaicML/{cls._MODEL}"

    @property
    def max_input_size(self) -> int:
        return self._CONTEXT_SIZE

    def _instructize_prompt(self, prompt: str, sources: list[Source]) -> str:
        # See https://huggingface.co/mosaicml/mpt-7b-instruct#formatting
        instruction = (
            "Use the following pieces of context to answer the question at the end. "
            "If you don't know the answer, just say so. Don't try to make up an answer.\n"
        )
        instruction += "\n\n".join(source.content for source in sources)
        return f"{instruction}### Instruction: {prompt}\n### Response:"

    def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> str:
        instruction = self._instructize_prompt(prompt, sources)
        # https://docs.mosaicml.com/en/latest/inference.html#text-completion-requests
        response = self._client.post(
            f"https://models.hosted-on.mosaicml.hosting/{self._MODEL}/v1/predict",
            headers={
                "Authorization": f"{self._api_key}",
                "Content-Type": "application/json",
            },
            json={
                "inputs": [instruction],
                "parameters": {"temperature": 0.0, "max_new_tokens": max_new_tokens},
            },
        )
        if response.is_error:
            raise RagnaException(
                status_code=response.status_code, response=response.json()
            )
        return cast(str, response.json()["outputs"][0]).replace(instruction, "").strip()


class Mpt7bInstruct(MosaicmlApiAssistant):
    """[MPT-7B-Instruct](https://docs.mosaicml.com/en/latest/inference.html#text-completion-models)

    !!! info "Required environment variables"

        - `MOSAICML_API_KEY`
    """

    # https://huggingface.co/mosaicml/mpt-7b-instruct#model-description
    _MODEL = "mpt-7b-instruct"
    _CONTEXT_SIZE = 2048


class Mpt30bInstruct(MosaicmlApiAssistant):
    """[MPT-30B-Instruct](https://docs.mosaicml.com/en/latest/inference.html#text-completion-models)

    !!! info "Required environment variables"

        - `MOSAICML_API_KEY`
    """

    # https://huggingface.co/mosaicml/mpt-30b-instruct#model-description
    _MODEL = "mpt-30b-instruct"
    _CONTEXT_SIZE = 8_192



import re
from datetime import datetime

import emoji
import httpx
import param


class RagnaAuthTokenExpiredException(Exception):
    """Just a wrapper around Exception"""

    pass


# The goal is this class is to provide ready-to-use functions to interact with the API
class ApiWrapper(param.Parameterized):
    auth_token = param.String(default=None)

    def __init__(self, api_url, **params):
        self.client = httpx.AsyncClient(base_url=api_url)

        super().__init__(**params)

        try:
            # If no auth token is provided, we use the API base URL and only test the API is up.
            # else, we test the API is up *and* the token is valid.
            endpoint = (
                api_url + "/components" if self.auth_token is not None else api_url
            )
            httpx.get(
                endpoint, headers={"Authorization": f"Bearer {self.auth_token}"}
            ).raise_for_status()

        except httpx.HTTPStatusError as e:
            # unauthorized - the token is invalid
            if e.response.status_code == 401:
                raise RagnaAuthTokenExpiredException("Unauthorized")
            else:
                raise e

    async def auth(self, username, password):
        self.auth_token = (
            (
                await self.client.post(
                    "/token",
                    data={"username": username, "password": password},
                )
            )
            .raise_for_status()
            .json()
        )

        return True

    @param.depends("auth_token", watch=True, on_init=True)
    def update_auth_header(self):
        self.client.headers["Authorization"] = f"Bearer {self.auth_token}"

    async def get_chats(self):
        json_data = (await self.client.get("/chats")).raise_for_status().json()
        for chat in json_data:
            chat["messages"] = [self.improve_message(msg) for msg in chat["messages"]]
        return json_data

    async def answer(self, chat_id, prompt):
        json_data = (
            (
                await self.client.post(
                    f"/chats/{chat_id}/answer",
                    params={"prompt": prompt},
                    timeout=None,
                )
            )
            .raise_for_status()
            .json()
        )

        json_data["message"] = self.improve_message(json_data["message"])
        json_data["chat"]["messages"] = [
            self.improve_message(msg) for msg in json_data["chat"]["messages"]
        ]

        return json_data

    async def get_components(self):
        return (await self.client.get("/components")).raise_for_status().json()

    # Upload and related functions
    def upload_endpoints(self):
        return {
            "informations_endpoint": f"{self.client.base_url}/document",
        }

    async def start_chat(self, name, documents, source_storage, assistant, params={}):
        return (
            (
                await self.client.post(
                    "/chats",
                    json={
                        "name": name,
                        "documents": documents,
                        "source_storage": source_storage,
                        "assistant": assistant,
                        "params": params,
                    },
                )
            )
            .raise_for_status()
            .json()
        )

    async def start_and_prepare(
        self, name, documents, source_storage, assistant, params={}
    ):
        chat = await self.start_chat(name, documents, source_storage, assistant, params)

        (
            await self.client.post(f"/chats/{chat['id']}/prepare", timeout=None)
        ).raise_for_status()

        return chat["id"]

    # Helpers

    def improve_message(self, msg):
        # convert timestamps to datetime

        msg["timestamp"] = datetime.strptime(msg["timestamp"], "%Y-%m-%dT%H:%M:%S.%f")

        msg["content"] = self.replace_emoji_shortcodes_with_emoji(msg["content"])

        return msg

    def replace_emoji_shortcodes_with_emoji(self, markdown_string):
        # Define a regular expression pattern to find emoji shortcodes
        shortcode_pattern = r":\w+:"

        # Find all matches of emoji shortcodes in the input string
        shortcodes = re.findall(shortcode_pattern, markdown_string)

        # Iterate through the found shortcodes and replace them with emojis
        for shortcode in shortcodes:
            emoji_name = shortcode.strip(":")
            emoji_unicode = emoji.emojize(f":{emoji_name}:", language="alias")
            markdown_string = markdown_string.replace(shortcode, emoji_unicode)

        return markdown_string



import itertools
from collections import defaultdict
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Annotated, Iterable, Type, TypeVar, cast

import pydantic
import questionary
import rich
import typer
from rich.table import Table

import ragna
from ragna.core import (
    Assistant,
    Config,
    EnvVarRequirement,
    PackageRequirement,
    RagnaException,
    Requirement,
    SourceStorage,
)
from ragna.core._components import Component


def parse_config(value: str) -> Config:
    try:
        config = Config.from_file(value)
    except RagnaException:
        rich.print(f"The configuration file {value} does not exist.")
        if value == "./ragna.toml":
            rich.print(
                "If you don't have a configuration file yet, "
                "run [bold]ragna init[/bold] to generate one."
            )
        raise typer.Exit(1)
    except pydantic.ValidationError as validation:
        # FIXME: pretty formatting!
        for error in validation.errors():
            rich.print(error)
        raise typer.Exit(1)
    # This stores the original value so we can pass it on to subprocesses that we might
    # start.
    config.__ragna_cli_config_path__ = value
    return config


ConfigOption = Annotated[
    ragna.Config,
    typer.Option(
        "-c",
        "--config",
        metavar="CONFIG",
        envvar="RAGNA_CONFIG",
        parser=parse_config,
        help="Path to a Ragna configuration file.",
    ),
]

# This adds a newline before every question to unclutter the output
QMARK = "\n?"


def init_config(*, output_path: Path, force: bool) -> tuple[Config, Path, bool]:
    # FIXME: add link to the config documentation when it is available
    rich.print(
        "\n\t[bold]Welcome to the Ragna config creation wizard![/bold]\n\n"
        "I'll help you create a configuration file to use with ragna.\n"
        "Due to the large amount of parameters, "
        "I unfortunately can't cover everything. "
        "If you want to customize everything, "
        "please have a look at the documentation instead."
    )

    intent = questionary.select(
        "Which of the following statements describes best what you want to do?",
        choices=[
            questionary.Choice(
                (
                    "I want to try Ragna without worrying about any additional "
                    "dependencies or setup."
                ),
                value="demo",
            ),
            questionary.Choice(
                (
                    "I want to try Ragna and its builtin source storages and "
                    "assistants, which potentially require additional dependencies "
                    "or setup."
                ),
                value="builtin",
            ),
            questionary.Choice(
                (
                    "I have used Ragna before and want to customize the most common "
                    "parameters."
                ),
                value="common",
            ),
        ],
        qmark=QMARK,
    ).unsafe_ask()

    wizard = {
        "demo": _wizard_demo,
        "builtin": _wizard_builtin,
        "common": _wizard_common,
    }[intent]
    config = wizard()

    if output_path.exists() and not force:
        output_path, force = _handle_output_path(output_path=output_path, force=force)

    rich.print(
        f"\nAnd with that we are done :tada: "
        f"I'm writing the configuration file to {output_path}."
    )

    return config, output_path, force


def _wizard_demo() -> Config:
    return Config()


def _wizard_builtin() -> Config:
    config = _wizard_demo()

    rich.print(
        "\nragna has the following components builtin. "
        "Select the ones that you want to use. "
        "If the requirements of a selected component are not met, "
        "I'll show you instructions how to meet them later."
    )
    config.core.source_storages = _select_components(
        "source storages",
        ragna.source_storages,
        SourceStorage,  # type: ignore[type-abstract]
    )
    config.core.assistants = _select_components(
        "assistants",
        ragna.assistants,
        Assistant,  # type: ignore[type-abstract]
    )

    _handle_unmet_requirements(
        itertools.chain(config.core.source_storages, config.core.assistants)
    )

    return config


T = TypeVar("T", bound=Component)


def _select_components(
    title: str,
    module: ModuleType,
    base_cls: Type[T],
) -> list[Type[T]]:
    components = [
        obj
        for obj in module.__dict__.values()
        if isinstance(obj, type) and issubclass(obj, base_cls) and obj is not base_cls
    ]
    return cast(
        list[Type[T]],
        questionary.checkbox(
            f"Which {title} do you want to use?",
            choices=[
                questionary.Choice(
                    component.display_name(),
                    value=component,
                    checked=component.is_available(),
                )
                for component in components
            ],
            qmark=QMARK,
        ).unsafe_ask(),
    )


def _handle_unmet_requirements(components: Iterable[Type[Component]]) -> None:
    unmet_requirements = set(
        requirement
        for component in components
        for requirement in component.requirements()
        if not requirement.is_available()
    )
    if not unmet_requirements:
        return

    rich.print(
        "You have selected components, which have additional requirements that are"
        "currently not met."
    )
    unmet_requirements_by_type = _split_requirements(unmet_requirements)

    unmet_package_requirements = sorted(
        str(requirement)
        for requirement in unmet_requirements_by_type[PackageRequirement]
    )
    if unmet_package_requirements:
        rich.print(
            "\nTo use the selected components, "
            "you need to install the following packages: \n"
        )
        for requirement in unmet_package_requirements:
            rich.print(f"- {requirement}")

        rich.print(
            f"\nTo do this, you can run\n\n"
            f"$ pip install {' '.join(unmet_package_requirements)}\n\n"
            f"Optionally, you can also install Ragna with all optional dependencies"
            f"for the builtin components\n\n"
            f"$ pip install 'ragna\[all]'"
        )

    unmet_env_var_requirements = sorted(
        str(requirement)
        for requirement in unmet_requirements_by_type[EnvVarRequirement]
    )
    if unmet_env_var_requirements:
        rich.print(
            "\nTo use the selected components, "
            "you need to set the following environment variables: \n"
        )
        for requirement in unmet_env_var_requirements:
            rich.print(f"- {requirement}")

    rich.print(
        "\nTip: You can check the availability of the requirements with "
        "[bold]ragna check[/bold]."
    )


def _wizard_common() -> Config:
    config = _wizard_builtin()

    config.local_cache_root = Path(
        questionary.path(
            "Where should local files be stored?",
            default=str(config.local_cache_root),
            qmark=QMARK,
        ).unsafe_ask()
    )

    config.core.queue_url = _select_queue_url(config)

    config.api.url = questionary.text(
        "At what URL do you want the ragna REST API to be served?",
        default=config.api.url,
        qmark=QMARK,
    ).unsafe_ask()

    if questionary.confirm(
        "Do you want to use a SQL database to persist the chats between runs?",
        default=True,
        qmark=QMARK,
    ).unsafe_ask():
        config.api.database_url = questionary.text(
            "What is the URL of the database?",
            default=f"sqlite:///{config.local_cache_root / 'ragna.db'}",
            qmark=QMARK,
        ).unsafe_ask()
    else:
        config.api.database_url = "memory"

    config.ui.url = questionary.text(
        "At what URL do you want the ragna web UI to be served?",
        default=config.ui.url,
        qmark=QMARK,
    ).unsafe_ask()

    return config


def _select_queue_url(config: Config) -> str:
    queue = questionary.select(
        (
            "Ragna internally uses a task queue to perform the RAG workflow. "
            "What kind of queue do you want to use?"
        ),
        # FIXME: include the descriptions as actual descriptions rather than as part
        #  of the title as soon as https://github.com/tmbo/questionary/issues/269 is
        #  resolved.
        choices=[
            questionary.Choice(
                (
                    "memory: Everything runs sequentially on the main thread "
                    "as if there were no task queue."
                ),
                value="memory",
            ),
            questionary.Choice(
                (
                    "file system: The local file system is used to build the queue. "
                    "Starting a ragna worker is required. "
                    "Requires the worker to be run on the same machine as the main "
                    "thread."
                ),
                value="file_system",
            ),
            questionary.Choice(
                (
                    "redis: Redis is used as queue. Starting a ragna worker is "
                    "required."
                ),
                value="redis",
            ),
        ],
        qmark=QMARK,
    ).unsafe_ask()

    if queue == "memory":
        return "memory"
    elif queue == "file_system":
        return cast(
            str,
            questionary.path(
                "Where do you want to store the queue files?",
                default=str(config.local_cache_root / "queue"),
                qmark=QMARK,
            ).unsafe_ask(),
        )
    else:  # queue == "redis":
        return cast(
            str,
            questionary.text(
                "What is the URL of the Redis instance?",
                default="redis://127.0.0.1:6379",
                qmark=QMARK,
            ).unsafe_ask(),
        )


def _handle_output_path(*, output_path: Path, force: bool) -> tuple[Path, bool]:
    rich.print(
        f"\nThe output path {output_path} already exists "
        f"and you didn't pass the --force flag to overwrite it. "
    )
    action = questionary.select(
        "What do you want to do?",
        choices=[
            questionary.Choice("Overwrite the existing file.", value="overwrite"),
            questionary.Choice("Select a new output path.", value="new"),
        ],
        qmark=QMARK,
    ).unsafe_ask()

    if action == "overwrite":
        force = True
    elif action == "new":
        while True:
            output_path = (
                Path(
                    questionary.path(
                        "Please provide a different output path "
                        "to write the generated config to:",
                        default=str(output_path),
                        qmark=QMARK,
                    ).unsafe_ask()
                )
                .expanduser()
                .resolve()
            )

            if not output_path.exists():
                break

            rich.print(f"The output path {output_path} already exists.")

    return output_path, force


def check_config(config: Config) -> bool:
    fully_available = True

    for title, components in [
        ("source storages", config.core.source_storages),
        ("assistants", config.core.assistants),
    ]:
        if TYPE_CHECKING:
            from ragna.core._components import Component

            components = cast(list[Type[Component]], components)

        table = Table(
            "",
            "name",
            "environment variables",
            "packages",
            show_lines=True,
            title=title,
        )

        for component in components:
            is_available = component.is_available()
            fully_available &= is_available

            requirements = _split_requirements(component.requirements())
            table.add_row(
                _yes_or_no(is_available),
                component.display_name(),
                _format_requirements(requirements[EnvVarRequirement]),
                _format_requirements(requirements[PackageRequirement]),
            )

        rich.print(table)

    return fully_available


def _split_requirements(
    requirements: Iterable[Requirement],
) -> dict[Type[Requirement], list[Requirement]]:
    split_reqs = defaultdict(list)
    for req in requirements:
        split_reqs[type(req)].append(req)
    return split_reqs


def _format_requirements(requirements: list[Requirement]) -> str:
    if not requirements:
        return ""

    return "\n".join(f"{_yes_or_no(req.is_available())} {req}" for req in requirements)


def _yes_or_no(condition: bool) -> str:
    return ":white_check_mark:" if condition else ":x:"



__all__ = [
    "Claude",
    "ClaudeInstant",
    "Gpt35Turbo16k",
    "Gpt4",
    "Mpt7bInstruct",
    "Mpt30bInstruct",
    "RagnaDemoAssistant",
]

from ._anthropic import Claude, ClaudeInstant
from ._demo import RagnaDemoAssistant
from ._mosaicml import Mpt7bInstruct, Mpt30bInstruct
from ._openai import Gpt4, Gpt35Turbo16k

# isort: split

from ragna._utils import fix_module

fix_module(globals())
del fix_module



import contextlib
import io
import json
import unittest.mock

import fastapi.openapi.utils
import mkdocs_gen_files
import typer.rich_utils

from ragna import Config
from ragna._api import app as api_app
from ragna._cli import app as cli_app


def main():
    cli_reference()
    api_reference()


def cli_reference():
    def get_help(command):
        with unittest.mock.patch.object(typer.rich_utils, "MAX_WIDTH", 80):
            with contextlib.suppress(SystemExit), contextlib.redirect_stdout(
                io.StringIO()
            ) as stdout:
                cli_app(([command] if command else []) + ["--help"], prog_name="ragna")

            return "\n".join(
                line.strip() for line in stdout.getvalue().strip().splitlines()
            )

    def get_doc(command):
        return (
            f"## `ragna{f' {command}' if command else ''}`\n\n"
            f"```\n{get_help(command)}\n```\n\n"
        )

    with mkdocs_gen_files.open("references/cli.md", "w") as file:
        file.write(f"# CLI reference\n\n{get_doc(None)}")
        for command in cli_app.registered_commands:
            file.write(get_doc(command.name or command.callback.__name__))


def api_reference():
    app = api_app(Config())
    openapi_json = fastapi.openapi.utils.get_openapi(
        title=app.title,
        version=app.version,
        openapi_version=app.openapi_version,
        description=app.description,
        routes=app.routes,
    )
    with mkdocs_gen_files.open("references/openapi.json", "w") as file:
        json.dump(openapi_json, file)


main()



import sys
from typing import Callable, Iterable, Iterator, Mapping, TypeVar

__all__ = ["itertools_pairwise", "importlib_metadata_package_distributions"]

T = TypeVar("T")


def _itertools_pairwise() -> Callable[[Iterable[T]], Iterator[tuple[T, T]]]:
    if sys.version_info[:2] >= (3, 10):
        from itertools import pairwise
    else:
        from itertools import tee
        from typing import Iterable, Iterator

        # https://docs.python.org/3/library/itertools.html#itertools.pairwise
        def pairwise(iterable: Iterable[T]) -> Iterator[tuple[T, T]]:
            # pairwise('ABCDEFG') --> AB BC CD DE EF FG
            a, b = tee(iterable)
            next(b, None)
            return zip(a, b)

    return pairwise


itertools_pairwise = _itertools_pairwise()


def _importlib_metadata_package_distributions() -> (
    Callable[[], Mapping[str, list[str]]]
):
    if sys.version_info[:2] >= (3, 10):
        from importlib.metadata import packages_distributions
    else:
        from importlib_metadata import packages_distributions

    return packages_distributions


importlib_metadata_package_distributions = _importlib_metadata_package_distributions()



import panel as pn
import param


class LogoutPage(pn.viewable.Viewer, param.Parameterized):
    def __init__(self, api_wrapper, **params):
        super().__init__(**params)
        self.api_wrapper = api_wrapper

        self.api_wrapper.auth_token = None

    def __panel__(self):
        # Usually, we do a redirect this way :
        # >>> pn.state.location.param.update(reload=True, pathname="/")
        # But it only works once the page is fully loaded.
        # So we render a javascript redirect instead.

        # To remove the token from the cookie, we have to force its expiry date to the past.
        return pn.pane.HTML(
            """<script>
                            document.cookie = "auth_token=; expires=Thu, 01 Jan 1970 00:00:00 UTC; path=/";
                            window.location.href = '/'; 
                            </script> """
        )



from typing import cast

from ragna.core import RagnaException, Source

from ._api import ApiAssistant


class OpenaiApiAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "OPENAI_API_KEY"
    _MODEL: str
    _CONTEXT_SIZE: int

    @classmethod
    def display_name(cls) -> str:
        return f"OpenAI/{cls._MODEL}"

    @property
    def max_input_size(self) -> int:
        return self._CONTEXT_SIZE

    def _make_system_content(self, sources: list[Source]) -> str:
        # See https://github.com/openai/openai-cookbook/blob/main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb
        instruction = (
            "You are an helpful assistants that answers user questions given the context below. "
            "If you don't know the answer, just say so. Don't try to make up an answer. "
            "Only use the sources below to generate the answer."
        )
        return instruction + "\n\n".join(source.content for source in sources)

    def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> str:
        # See https://platform.openai.com/docs/api-reference/chat/create
        # and https://platform.openai.com/docs/api-reference/chat/object
        response = self._client.post(
            "https://api.openai.com/v1/chat/completions",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self._api_key}",
            },
            json={
                "messages": [
                    {
                        "role": "system",
                        "content": self._make_system_content(sources),
                    },
                    {
                        "role": "user",
                        "content": prompt,
                    },
                ],
                "model": self._MODEL,
                "temperature": 0.0,
                "max_tokens": max_new_tokens,
            },
        )
        if response.is_error:
            raise RagnaException(
                status_code=response.status_code, response=response.json()
            )
        return cast(str, response.json()["choices"][0]["message"]["content"])


class Gpt35Turbo16k(OpenaiApiAssistant):
    """[OpenAI GPT-3.5](https://platform.openai.com/docs/models/gpt-3-5)

    !!! info "Required environment variables"

        - `OPENAI_API_KEY`
    """

    _MODEL = "gpt-3.5-turbo-16k"
    _CONTEXT_SIZE = 16_384


Gpt35Turbo16k.__doc__ = "OOPS"


class Gpt4(OpenaiApiAssistant):
    """[OpenAI GPT-4](https://platform.openai.com/docs/models/gpt-4)

    !!! info "Required environment variables"

        - `OPENAI_API_KEY`
    """

    _MODEL = "gpt-4"
    _CONTEXT_SIZE = 8_192



import re
import sys
import textwrap

from ragna.core import Assistant, Source


class RagnaDemoAssistant(Assistant):
    """Demo assistant without requirements.

    !!! note

        As the name implies, this assistant is just for demo purposes and cannot
        answer any questions. By default it replies with the prompt and the given
        sources.

        If you include the phrase `"markdown"` into your prompt, it will return a
        Markdown table including emojis.
    """

    @classmethod
    def display_name(cls) -> str:
        return "Ragna/DemoAssistant"

    @property
    def max_input_size(self) -> int:
        return sys.maxsize

    def answer(self, prompt: str, sources: list[Source]) -> str:
        if re.search("markdown", prompt, re.IGNORECASE):
            return self._markdown_answer()
        else:
            return self._default_answer(prompt, sources)

    def _markdown_answer(self) -> str:
        return textwrap.dedent(
            """
            | String | Integer  | Float | Emoji              |
            | :----- | :------: | ----: | ------------------ |
            | foo    | 0        | 1.0   | :unicorn:          |
            | `bar`  | 1        | -1.23 | :metal:            |
            | "baz"  | -1       | 1e6   | :eye: :lips: :eye: |
            """
        ).strip()

    def _default_answer(self, prompt: str, sources: list[Source]) -> str:
        sources_display = []
        for source in sources:
            source_display = f"- {source.document.name}"
            if source.location:
                source_display += f", {source.location}"
            source_display += f": {textwrap.shorten(source.content, width=100)}"
            sources_display.append(source_display)
        if len(sources) > 3:
            sources_display.append("[...]")

        return (
            textwrap.dedent(
                """
                I'm a demo assistant and can be used to try Ragnas workflow. I will only
                mirror back my inputs. 
                
                Your prompt was:

                > {prompt}

                These are the sources I was given:

                {sources}
                """
            )
            .strip()
            .format(name=str(self), prompt=prompt, sources="\n".join(sources_display))
        )
